/* bmc
 * read "Symbolic Model Checking using SAT procedures instead of BDDs"
 * to get the stuff
 *--
 * public domain
 * may 2002
 */

#include "bmc.h"

#include <stdio.h>
#include <stdlib.h>
#include <string.h>

#include "trans.h"
#include "vars.h"
#include "bdd.h"
#include "verbose.h"
#include "checker.h"
#include "ctl.h"
#include "mem.h"

/* return f with weight = in for input vars and weight = out for output vars */
/* what to do with input and output ? currently not handled */
static transition_tree copy_rename(trans_data *d, variables *v, transition_tree f, int in, int out)
{
  switch (get_transition_tree_type(d, f)) {
    case CTL_TRUE :
    case CTL_FALSE :
      return f;
    case VAR : {
      switch (get_variable_type(v, get_transition_tree_variable(d, f))) {
	case VARIABLE_TYPE_CURRENT :
	  return new_transition_tree_var2(d, get_variable_canon(v, get_transition_tree_variable(d, f)), in);
	case VARIABLE_TYPE_NEXT :
	  return new_transition_tree_var2(d, get_variable_canon(v, get_transition_tree_variable(d, f)), out);
	default :
	  fprintf(stderr, "%s:%d:%s: error : unhandled type %d\n", __FILE__, __LINE__, __FUNCTION__, get_transition_tree_type(d, f));
	  abort();
      }
    }
    case NOT :
      return new_transition_tree_op(d, NOT, copy_rename(d, v, get_transition_tree_son(d, f), in, out), TRANS_UNDEF);
    case AND :
    case OR :
    case XOR :
    case EQUAL :
    case IMPLIES :
      return new_transition_tree_op(d, get_transition_tree_type(d, f),
		copy_rename(d, v, get_transition_tree_left_son(d, f), in, out),
		copy_rename(d, v, get_transition_tree_right_son(d, f), in, out));
    default :
      fprintf(stderr, "%s:%d:%s: error : unhandled type %d\n", __FILE__, __LINE__, __FUNCTION__, get_transition_tree_type(d, f));
      abort();

  }
}

transition_tree Mk(trans_data *d, variables *v, transition_tree fonction, transition_tree init, int k)
{
  int i;
  transition_tree tmp, ret;

  transition_tree s0 = copy_rename(d, v, init, 0, -1);

  ret = s0;

  print("bmc", stdout, "k = %d\n", k);

  for (i=0; i<k; i++) {
    tmp = copy_rename(d, v, fonction, i, i+1);

    if (verbose_on("bmc")) {
      extern checker the_checker;
      fprintf(stdout, "i = %d\n", i);
      print_modified_tree(d, v, tmp, the_checker.vars);
      fprintf(stdout, "\n");
    }

    ret = new_transition_tree_op(d, AND, ret, tmp);
  }

  return ret;
}

transition_tree lLk(trans_data *d, variables *v, transition_tree function, int k, int l)
{
  return copy_rename(d, v, function, k, l);
}

transition_tree Lk(trans_data *d, variables *v, transition_tree function, int k)
{
  int l;
  transition_tree ret;

  ret = lLk(d, v, function, k, 0);

  for (l=1; l<k+1; l++)
    ret = new_transition_tree_op(d, OR, ret, lLk(d, v, function, k, l));

  if (verbose_on("bmc")) {
    extern checker the_checker;
    fprintf(stdout, "Lk for k = %d\n", k);
    print_modified_tree(d, v, ret, the_checker.vars);
    fprintf(stdout, "\n");
  }

  return ret;
}

transition_tree translate_ltl(trans_data *d, variables *v, ctl_data *c, ctl_formula f, int k, int i)
{
  switch (get_ctl_type(c, f)) {
    case CTL_TRUE :
      return new_transition_tree_true(d);

    case CTL_FALSE :
      return new_transition_tree_false(d);

    case ATOM :
      return new_transition_tree_var2(d, get_variable_canon(v, get_ctl_variable(c, f)), i);

    case NOT : {
      transition_tree a = translate_ltl(d, v, c, get_ctl_son(c, f), k, i);
      if (get_transition_tree_type(d, a) != VAR) {
	fprintf(stderr, "%s:%d:%s: error : unhandled type %d\n", __FILE__, __LINE__, __FUNCTION__, get_transition_tree_type(d, a));
	abort();
      }
      return new_transition_tree_op(d, NOT, a, TRANS_UNDEF);
    }

    case ALWAYS :
      return new_transition_tree_false(d);

    case EVENTUALLY : {
      transition_tree ret;
      int j;

      if (i>k) {
	fprintf(stderr, "%s:%d:%s: error : i>k (i=%d k=%d)\n", __FILE__, __LINE__, __FUNCTION__, i, k);
	abort();
      }

      ret = translate_ltl(d, v, c, get_ctl_son(c, f), k, i);

      for (j=i+1; j<=k; j++)
	ret = new_transition_tree_op(d, OR, ret, translate_ltl(d, v, c, get_ctl_son(c, f), k, j));

      return ret;
    }

    case NEXT : {
      if (i>=k)
	return new_transition_tree_false(d);

      return translate_ltl(d, v, c, get_ctl_son(c, f), k, i+1);
    }

    case UNTIL : {
      int j, n;
      ctl_formula ff = get_ctl_left_son(c, f);
      ctl_formula g = get_ctl_right_son(c, f);
      transition_tree ret=TRANS_UNDEF;

      for (j=i; j<=k; j++) {
	transition_tree cur;

        cur = translate_ltl(d, v, c, g, k, j);

	for (n=i; n<=j-1; n++)
	  cur = new_transition_tree_op(d, AND, cur, translate_ltl(d, v, c, ff, k, n));

	if (ret == TRANS_UNDEF)
	  ret = cur;
	else
	  ret = new_transition_tree_op(d, OR, ret, cur);
      }

      if (ret == TRANS_UNDEF) {
	fprintf(stderr, "%s:%d:%simpossible case ?\n", __FILE__, __LINE__, __FUNCTION__);
	abort();
      }

      return ret;
    }

    case AND : {
      transition_tree a = translate_ltl(d, v, c, get_ctl_left_son(c, f), k, i);
      transition_tree b = translate_ltl(d, v, c, get_ctl_right_son(c, f), k, i);

      return new_transition_tree_op(d, AND, a, b);
    }

    case OR : {
      transition_tree a = translate_ltl(d, v, c, get_ctl_left_son(c, f), k, i);
      transition_tree b = translate_ltl(d, v, c, get_ctl_right_son(c, f), k, i);

      return new_transition_tree_op(d, OR, a, b);
    }

    default :
      fprintf(stderr, "%s:%d:%s: error : unhandled type %d\n", __FILE__, __LINE__, __FUNCTION__, get_ctl_type(c, f));
      abort();

  }
}

static int fatal_error(void)
{
  fprintf(stderr, "error, i>k, impossible case\n");
  abort();

  /* gcc doesn't complain, but it looks strange to me without it */
  return 0;
}

transition_tree translate_ltl_loop(trans_data *d, variables *v, ctl_data *c, ctl_formula f, int k, int l, int i)
{
#define min(i, l) ((i) < (l) ? (i) : (l))
#define succ(i) ((i) < k ? (i) + 1 : ((i) == k ? l : fatal_error()))

  switch (get_ctl_type(c, f)) {
    case CTL_TRUE :
      return new_transition_tree_true(d);

    case CTL_FALSE :
      return new_transition_tree_false(d);

    case ATOM :
      return new_transition_tree_var2(d, get_variable_canon(v, get_ctl_variable(c, f)), i);

    case NOT : {
      transition_tree a = translate_ltl_loop(d, v, c, get_ctl_son(c, f), k, l, i);
      if (get_transition_tree_type(d, a) != VAR) {
	fprintf(stderr, "%s:%d:%s: error : unhandled type %d\n", __FILE__, __LINE__, __FUNCTION__, get_transition_tree_type(d, a));
	abort();
      }
      return new_transition_tree_op(d, NOT, a, TRANS_UNDEF);
    }

    case ALWAYS : {
      int j;
      transition_tree ret = TRANS_UNDEF;

      for (j = min(i, l); j<=k; j++) {
	transition_tree cur;

	cur = translate_ltl_loop(d, v, c, get_ctl_son(c, f), k, l, j);

	if (ret == TRANS_UNDEF)
	  ret = cur;
	else
	  ret = new_transition_tree_op(d, AND, ret, cur);
      }

      if (ret == TRANS_UNDEF) {
	fprintf(stderr, "%s:%d:%simpossible case ?\n", __FILE__, __LINE__, __FUNCTION__);
	abort();
      }

      return ret;
    }

    case EVENTUALLY : {
      int j;
      transition_tree ret = TRANS_UNDEF;

      for (j = min(i, l); j<=k; j++) {
	transition_tree cur;

	cur = translate_ltl_loop(d, v, c, get_ctl_son(c, f), k, l, j);

	if (ret == TRANS_UNDEF)
	  ret = cur;
	else
	  ret = new_transition_tree_op(d, OR, ret, cur);
      }

      if (ret == TRANS_UNDEF) {
	fprintf(stderr, "%s:%d:%simpossible case ?\n", __FILE__, __LINE__, __FUNCTION__);
	abort();
      }

      return ret;
    }

    case NEXT :
      return translate_ltl_loop(d, v, c, get_ctl_son(c, f), k, l, succ(i));

    case UNTIL : {
      transition_tree ff = get_ctl_left_son(c, f);
      transition_tree g = get_ctl_right_son(c, f);
      transition_tree t1=TRANS_UNDEF, t2=TRANS_UNDEF;
      int j, n;

      for (j = i; j <= k; j++) {
	transition_tree cur;

	cur = translate_ltl_loop(d, v, c, g, k, l, j);

	for (n = i; n <= j-1; n++)
	  cur = new_transition_tree_op(d, AND, cur, translate_ltl_loop(d, v, c, ff, k, l, n));

	if (t1 == TRANS_UNDEF)
	  t1 = cur;
	else
	  t1 = new_transition_tree_op(d, OR, t1, cur);
      }

      for (j = l; j <= i-1; j++) {
	transition_tree cur;

	cur = translate_ltl_loop(d, v, c, g, k, l, j);

	for (n = i; n <= k; n++)
	  cur = new_transition_tree_op(d, AND, cur, translate_ltl_loop(d, v, c, ff, k, l, n));

	for (n = l; n <= j-1; n++)
	  cur = new_transition_tree_op(d, AND, cur, translate_ltl_loop(d, v, c, ff, k, l, n));

	if (t2 == TRANS_UNDEF)
	  t2 = cur;
	else
	  t2 = new_transition_tree_op(d, OR, t2, cur);
      }

      if (t1 == TRANS_UNDEF && t2 == TRANS_UNDEF) {
	fprintf(stderr, "%s:%d:%s:impossible case ?\n", __FILE__, __LINE__, __FUNCTION__);
	abort();
      }

      if (t1 == TRANS_UNDEF)
	return t2;

      if (t2 == TRANS_UNDEF)
	return t1;

      return new_transition_tree_op(d, OR, t1, t2);
    }

    case AND : {
      transition_tree a = translate_ltl_loop(d, v, c, get_ctl_left_son(c, f), k, l, i);
      transition_tree b = translate_ltl_loop(d, v, c, get_ctl_right_son(c, f), k, l, i);

      return new_transition_tree_op(d, AND, a, b);
    }

    case OR : {
      transition_tree a = translate_ltl_loop(d, v, c, get_ctl_left_son(c, f), k, l, i);
      transition_tree b = translate_ltl_loop(d, v, c, get_ctl_right_son(c, f), k, l, i);

      return new_transition_tree_op(d, OR, a, b);
    }

    default :
      fprintf(stderr, "%s:%d:%s: error : unhandled type %d\n", __FILE__, __LINE__, __FUNCTION__, get_ctl_type(c, f));
      abort();
  }
}

void print_modified_tree(trans_data *d, variables *v, transition_tree t, bdd_variables_set *s)
{
  static int current = -1;
  int pcur = current;

  if (trans_undef(t)) {
    fprintf(stdout, "(transition tree undefined)");
    return;
  }

  current = get_transition_tree_type(d, t);

  switch (current) {
    case VAR :
#if 0
      fprintf(stdout, "(%s:%d)", get_variable_name(v, s->list2[get_transition_tree_variable(d, t)]),
			       get_transition_tree_variable_pos(d, t));
#endif
      fprintf(stdout, "(%d)", get_transition_tree_variable(d, t));
      break;
    case CTL_TRUE :
      fprintf(stdout, "true");
      break;
    case CTL_FALSE :
      fprintf(stdout, "false");
      break;
    case OR :
    case XOR :
      if (pcur == AND || pcur == NOT)
        fprintf(stdout, "(");
      print_modified_tree(d, v, get_transition_tree_left_son(d, t), s);
      fprintf(stdout, "%c", current == OR ? '+' : '^');
      print_modified_tree(d, v, get_transition_tree_right_son(d, t), s);
      if (pcur == AND || pcur == NOT)
        fprintf(stdout, ")");
      break;
    case AND :
      if (pcur == NOT)
        fprintf(stdout, "(");
      print_modified_tree(d, v, get_transition_tree_left_son(d, t), s);
      fprintf(stdout, ".");
      print_modified_tree(d, v, get_transition_tree_right_son(d, t), s);
      if (pcur == NOT)
        fprintf(stdout, ")");
      break;
    case IMPLIES :
      if (pcur != -1)
        fprintf(stdout, "(");
      print_modified_tree(d, v, get_transition_tree_left_son(d, t), s);
      fprintf(stdout, "=>");
      print_modified_tree(d, v, get_transition_tree_right_son(d, t), s);
      if (pcur != -1)
        fprintf(stdout, ")");
      break;
    case EQUAL :
      if (pcur != -1)
        fprintf(stdout, "(");
      print_modified_tree(d, v, get_transition_tree_left_son(d, t), s);
      fprintf(stdout, "=");
      print_modified_tree(d, v, get_transition_tree_right_son(d, t), s);
      if (pcur != -1)
        fprintf(stdout, ")");
      break;
    case NOT :
      fprintf(stdout, "!");
      print_modified_tree(d, v, get_transition_tree_son(d, t), s);
      break;
    default :
      fprintf(stdout, "(unknown type %d)", current);
      break;
  }

  current = pcur;
}

static ctl_formula down_not(ctl_data *, ctl_formula);

static ctl_formula notify(ctl_data *d, ctl_formula f)
{
  switch (get_ctl_type(d, f)) {
    case ATOM :
      return new_ctl_formula_op(d, NOT, f, CTL_NO_FORMULA);
    case CTL_TRUE :
      return new_ctl_formula_false(d);
    case CTL_FALSE :
      return new_ctl_formula_true(d);
    case ALWAYS : {
      /* not always x = eventually not x */
      ctl_formula x = get_ctl_son(d, f);
      x = down_not(d, new_ctl_formula_op(d, NOT, x, CTL_NO_FORMULA));
      return new_ctl_formula_op(d, EVENTUALLY, x, CTL_NO_FORMULA);
    }
    case EVENTUALLY : {
      /* not eventually x = always not x */
      ctl_formula x = get_ctl_son(d, f);
      x = down_not(d, new_ctl_formula_op(d, NOT, x, CTL_NO_FORMULA));
      return new_ctl_formula_op(d, ALWAYS, x, CTL_NO_FORMULA);
    }
    case NEXT : {
      /* not next x = next not x */
      ctl_formula x = get_ctl_son(d, f);
      x = down_not(d, new_ctl_formula_op(d, NOT, x, CTL_NO_FORMULA));
      return new_ctl_formula_op(d, NEXT, x, CTL_NO_FORMULA);
    }
    case UNTIL : {
      /* ~(a U b) = (~b U (~a & ~b)) | ~<> b */
      /* not a until b = (not b U (not a and not b)) or always not b*/
      ctl_formula a = get_ctl_left_son(d, f);
      ctl_formula b = get_ctl_right_son(d, f);
      ctl_formula na, nb, naandnb, u, anb;
      na = down_not(d, new_ctl_formula_op(d, NOT, a, CTL_NO_FORMULA));
      nb = down_not(d, new_ctl_formula_op(d, NOT, b, CTL_NO_FORMULA));
      naandnb = new_ctl_formula_op(d, AND, na, nb);
      u = new_ctl_formula_op(d, UNTIL, nb, naandnb);
      anb = new_ctl_formula_op(d, ALWAYS, nb, CTL_NO_FORMULA);
      return new_ctl_formula_op(d, OR, u, anb);
    }
    case AND : {
      /* not a and b = !a + !b */
      ctl_formula a = get_ctl_left_son(d, f);
      ctl_formula b = get_ctl_right_son(d, f);
      ctl_formula na, nb;
      na = down_not(d, new_ctl_formula_op(d, NOT, a, CTL_NO_FORMULA));
      nb = down_not(d, new_ctl_formula_op(d, NOT, b, CTL_NO_FORMULA));
      return new_ctl_formula_op(d, OR, na, nb);
    }
    case OR : {
      /* not a or b = !a . !b */
      ctl_formula a = get_ctl_left_son(d, f);
      ctl_formula b = get_ctl_right_son(d, f);
      ctl_formula na, nb;
      na = down_not(d, new_ctl_formula_op(d, NOT, a, CTL_NO_FORMULA));
      nb = down_not(d, new_ctl_formula_op(d, NOT, b, CTL_NO_FORMULA));
      return new_ctl_formula_op(d, AND, na, nb);
    }
    case XOR : {
      /* not a xor b = !(a.!b + !a.b) = (!a+b).(a+!b) = a.b + !a.!b */
      ctl_formula a = down_not(d, get_ctl_left_son(d, f));
      ctl_formula b = down_not(d, get_ctl_right_son(d, f));
      ctl_formula na, nb, ad, nad;
      na = down_not(d, new_ctl_formula_op(d, NOT, a, CTL_NO_FORMULA));
      nb = down_not(d, new_ctl_formula_op(d, NOT, b, CTL_NO_FORMULA));
      ad = new_ctl_formula_op(d, AND, a, b);
      nad = new_ctl_formula_op(d, AND, na, nb);
      return new_ctl_formula_op(d, OR, ad, nad);
    }
    case IMPLIES : {
      /* not a => b = not (!a+b) = a.!b */
      ctl_formula a = down_not(d, get_ctl_left_son(d, f));
      ctl_formula b = get_ctl_right_son(d, f);
      ctl_formula nb;
      nb = down_not(d, new_ctl_formula_op(d, NOT, b, CTL_NO_FORMULA));
      return new_ctl_formula_op(d, AND, a, nb);
    }
    case NOT :
      /* not not x = x */
      return down_not(d, get_ctl_son(d, f));
    default :
      fprintf(stderr, "%s:%d:%s: error : unhandled type %d\n", __FILE__, __LINE__, __FUNCTION__, get_ctl_type(d, f));
      abort();
  }
}

static ctl_formula down_not(ctl_data *d, ctl_formula f)
{
  switch (get_ctl_type(d, f)) {
    case ATOM :
    case CTL_TRUE :
    case CTL_FALSE :
      return f;
    case ALWAYS :
    case EVENTUALLY :
    case NEXT :
      return new_ctl_formula_op(d, get_ctl_type(d, f), down_not(d, get_ctl_son(d, f)), CTL_NO_FORMULA);
    case UNTIL :
    case AND :
    case OR :
      return new_ctl_formula_op(d, get_ctl_type(d, f), down_not(d, get_ctl_left_son(d, f)), down_not(d, get_ctl_right_son(d, f)));
    case XOR : { /* a xor b = !a.b + a.!b */
      ctl_formula a = down_not(d, get_ctl_left_son(d, f));
      ctl_formula b = down_not(d, get_ctl_right_son(d, f));
      ctl_formula na, nb, a1, a2;
      na = down_not(d, new_ctl_formula_op(d, NOT, a, CTL_NO_FORMULA));
      nb = down_not(d, new_ctl_formula_op(d, NOT, b, CTL_NO_FORMULA));
      a1 = new_ctl_formula_op(d, AND, na, b);
      a2 = new_ctl_formula_op(d, AND, a, nb);
      return new_ctl_formula_op(d, OR, a1, a2);
    }
    case IMPLIES : { /* a => b = !a+b */
      ctl_formula a = get_ctl_left_son(d, f);
      ctl_formula b = down_not(d, get_ctl_right_son(d, f));
      ctl_formula na;
      na = down_not(d, new_ctl_formula_op(d, NOT, a, CTL_NO_FORMULA));
      return new_ctl_formula_op(d, OR, na, b);
    }
    case NOT :
      return notify(d, get_ctl_son(d, f));
    default :
      fprintf(stderr, "%s:%d:%s: error : unhandled type %d\n", __FILE__, __LINE__, __FUNCTION__, get_ctl_type(d, f));
      abort();
  }
}

ctl_formula negate_ltl_formula(ctl_data *d, ctl_formula f)
{
  ctl_formula ret = new_ctl_formula_op(d, NOT, f, CTL_NO_FORMULA);

  if (verbose_on("bmc")) {
    extern checker the_checker;
    fprintf(stdout, "ltl formula before negation :\n");
    print_formula(&the_checker.ctl_data, &the_checker.variables, f);
    fprintf(stdout, "\n");
  }

  ret = down_not(d, ret);

  if (verbose_on("bmc")) {
    extern checker the_checker;
    fprintf(stdout, "ltl formula after negation :\n");
    print_formula(&the_checker.ctl_data, &the_checker.variables, ret);
    fprintf(stdout, "\n");
  }

  return ret;
}

/********************************************************************/
/*                                                                  */
/* to_cnf2, the real one, creates variables on the fly, on per node */
/*                                                                  */
/********************************************************************/
static int nn;
static char *vars;
static long vars_size;


static void numberize(trans_data *d, transition_tree t)
{
  if (d->trans_tree[t].number)
    return;

  d->trans_tree[t].number = nn;
  nn++;

  switch(get_transition_tree_type(d, t)) {
    case VAR :
    case CTL_TRUE :
    case CTL_FALSE :
      return;
    case NOT :
      numberize(d, get_transition_tree_son(d, t));
      return;
    case AND :
    case OR :
    case XOR :
    case EQUAL :
    case IMPLIES :
      numberize(d, get_transition_tree_left_son(d, t));
      numberize(d, get_transition_tree_right_son(d, t));
      return;
    default :
      fprintf(stderr, "trans.c:%d:numberize: error : unhandled type %d\n", __LINE__, get_transition_tree_type(d, t));
      abort();
  }
}

/* let's numberize leaves */
static void numberize2(trans_data *d, transition_tree t)
{
  if (d->trans_tree[t].number)
    return;

  switch(get_transition_tree_type(d, t)) {
    case VAR :
      d->trans_tree[t].number = get_transition_tree_variable(d, t);
      nn++;
      return;
    case CTL_TRUE :
    case CTL_FALSE :
      return;
    case NOT :
      numberize2(d, get_transition_tree_son(d, t));
      return;
    case AND :
    case OR :
    case XOR :
    case EQUAL :
    case IMPLIES :
      numberize2(d, get_transition_tree_left_son(d, t));
      numberize2(d, get_transition_tree_right_son(d, t));
      return;
    default :
      fprintf(stderr, "trans.c:%d:numberize: error : unhandled type %d\n", __LINE__, get_transition_tree_type(d, t));
      abort();
  }
}

static void closify(clauses *cl, trans_data *d, transition_tree t)
{
#if 0
transition_tree t1, t2, t3;
t1 = new_transition_tree_var(d, 10);
t2 = new_transition_tree_var(d, 11);

d->trans_tree[t1].number = 10;
d->trans_tree[t2].number = 11;
t3 = new_transition_tree_op(d, IMPLIES, t1, t2);
d->trans_tree[t3].number = 12;
add_clause(nd, c, d, t3);
return;
#endif

  if (d->trans_tree[t].added == 1)
    return;

  d->trans_tree[t].added = 1;

  switch(get_transition_tree_type(d, t)) {
    case VAR :
    case CTL_TRUE :
    case CTL_FALSE :
      return;
    case NOT : { /* a = !b == (a + b) . (!a + !b) */
      unsigned long a = d->trans_tree[t].number;
      unsigned long b = d->trans_tree[get_transition_tree_son(d, t)].number;
      clause *l;

      l = new_clause(cl);
      add_atom(l, a);
      add_atom(l, b);

      l = new_clause(cl);
      add_atom(l, -a);
      add_atom(l, -b);

      return closify(cl, d, get_transition_tree_son(d, t));
    }

    case AND : { /* a = b . c == (a + !b + !c) . (!a + b) . (!a + c) */
      unsigned long a = d->trans_tree[t].number;
      unsigned long b = d->trans_tree[get_transition_tree_left_son(d, t)].number;
      unsigned long c = d->trans_tree[get_transition_tree_right_son(d, t)].number;
      clause *l;

      l= new_clause(cl);
      add_atom(l, a);
      add_atom(l, -b);
      add_atom(l, -c);

      l= new_clause(cl);
      add_atom(l, -a);
      add_atom(l, b);

      l= new_clause(cl);
      add_atom(l, -a);
      add_atom(l, c);

      closify(cl, d, get_transition_tree_left_son(d, t));
      return closify(cl, d, get_transition_tree_right_son(d, t));
    }

    case OR : { /* a = b + c == (!a + b + c) . (a + !b) . (a + !c) */
      unsigned long a = d->trans_tree[t].number;
      unsigned long b = d->trans_tree[get_transition_tree_left_son(d, t)].number;
      unsigned long c = d->trans_tree[get_transition_tree_right_son(d, t)].number;
      clause *l;

      l= new_clause(cl);
      add_atom(l, -a);
      add_atom(l, b);
      add_atom(l, c);

      l= new_clause(cl);
      add_atom(l, a);
      add_atom(l, -b);

      l= new_clause(cl);
      add_atom(l, a);
      add_atom(l, -c);

      closify(cl, d, get_transition_tree_left_son(d, t));
      return closify(cl, d, get_transition_tree_right_son(d, t));
    }

    case XOR : { /* a = b xor c == (a + !b + c) . (a + b + !c) . (!a + b + c) . (!a + !b + !c) */
      unsigned long a = d->trans_tree[t].number;
      unsigned long b = d->trans_tree[get_transition_tree_left_son(d, t)].number;
      unsigned long c = d->trans_tree[get_transition_tree_right_son(d, t)].number;
      clause *l;

      l= new_clause(cl);
      add_atom(l, a);
      add_atom(l, -b);
      add_atom(l, c);

      l= new_clause(cl);
      add_atom(l, a);
      add_atom(l, b);
      add_atom(l, -c);

      l= new_clause(cl);
      add_atom(l, -a);
      add_atom(l, b);
      add_atom(l, c);

      l= new_clause(cl);
      add_atom(l, -a);
      add_atom(l, -b);
      add_atom(l, -c);

      closify(cl, d, get_transition_tree_left_son(d, t));
      return closify(cl, d, get_transition_tree_right_son(d, t));
    }

    case EQUAL : { /* a = (b = c) == (a + !b + !c) . (a + b + c) . (!a + !b + c) . (!a + b + !c) */
      unsigned long a = d->trans_tree[t].number;
      unsigned long b = d->trans_tree[get_transition_tree_left_son(d, t)].number;
      unsigned long c = d->trans_tree[get_transition_tree_right_son(d, t)].number;
      clause *l;

      l= new_clause(cl);
      add_atom(l, a);
      add_atom(l, -b);
      add_atom(l, -c);

      l= new_clause(cl);
      add_atom(l, a);
      add_atom(l, b);
      add_atom(l, c);

      l= new_clause(cl);
      add_atom(l, -a);
      add_atom(l, -b);
      add_atom(l, c);

      l= new_clause(cl);
      add_atom(l, -a);
      add_atom(l, b);
      add_atom(l, -c);

      closify(cl, d, get_transition_tree_left_son(d, t));
      return closify(cl, d, get_transition_tree_right_son(d, t));
    }

    case IMPLIES : { /* a = (b => c) == (a + b) . (a + !c) . (!a . !b . c) */
      unsigned long a = d->trans_tree[t].number;
      unsigned long b = d->trans_tree[get_transition_tree_left_son(d, t)].number;
      unsigned long c = d->trans_tree[get_transition_tree_right_son(d, t)].number;
      clause *l;

      l= new_clause(cl);
      add_atom(l, a);
      add_atom(l, b);

      l= new_clause(cl);
      add_atom(l, a);
      add_atom(l, -c);

      l= new_clause(cl);
      add_atom(l, -a);
      add_atom(l, -b);
      add_atom(l, c);

      closify(cl, d, get_transition_tree_left_son(d, t));
      return closify(cl, d, get_transition_tree_right_son(d, t));
    }

    default :
      fprintf(stderr, "trans.c:%d:closify: error : unhandled type %d\n", __LINE__, get_transition_tree_type(d, t));
      abort();
  }
}

/* returns the number of variables created */
unsigned long to_cnf(clauses *c, trans_data *d, transition_tree t)
{
  extern checker the_checker;

  print("cnf", stdout, "initial transition function :\n");
  if (verbose_on("cnf")) {
    print_modified_tree(d, &the_checker.variables, t, the_checker.vars);
    fprintf(stdout, "\n");
  }

  nn = 1;
  numberize2(d, t);
  numberize(d, t);

  vars = getmem(nn * sizeof(char));
  vars_size = nn;

  memset(vars, 0, vars_size * sizeof(char));

  closify(c, d, t);

  freemem(vars);

  return nn-1;
}

void dump_variable_names(trans_data *d, transition_tree t, variables *v, bdd_variables_set *s)
{
  if (d->trans_tree[t].added == 0)
    return;
  d->trans_tree[t].added = 0;

  switch(get_transition_tree_type(d, t)) {
    case CTL_TRUE :
      fprintf(stdout, "c %d = Strue[0]\n", d->trans_tree[t].number);
      return;
    case CTL_FALSE :
      fprintf(stdout, "c %d = Sfalse[0]\n", d->trans_tree[t].number);
      return;
    case VAR :
#if 0
      fprintf(stdout, "c %d = V%d_%s[0]\n", d->trans_tree[t].number, get_transition_tree_variable_pos(d, t),
					    get_variable_name(v, s->list2[get_transition_tree_variable(d, t)]));
#endif
      fprintf(stdout, "c %d = V%d\n", d->trans_tree[t].number, get_transition_tree_variable(d, t));
      return;
    case NOT :
      dump_variable_names(d, get_transition_tree_son(d, t), v, s);
      return;
    case AND :
    case OR :
    case XOR :
    case IMPLIES :
    case EQUAL :
      dump_variable_names(d, get_transition_tree_left_son(d, t), v, s);
      dump_variable_names(d, get_transition_tree_right_son(d, t), v, s);
      return;
  }
}

/* special cases handling (those implemented into bmc, very fast in fact) */

/* return 0 if f contains no X, U, G, F */
/* could be optimized (no recursion or longjmp) */
int no_temporal(ctl_data *d, ctl_formula f)
{
  switch(get_ctl_type(d, f)) {
    case ALWAYS :
    case EVENTUALLY :
    case NEXT :
    case UNTIL :
      return 0;
    case CTL_TRUE :
    case CTL_FALSE :
    case ATOM :
      return 1;
    case NOT :
      return no_temporal(d, get_ctl_son(d, f));
    case AND :
    case OR :
    case XOR :
    case EQUAL :
    case IMPLIES :
      return no_temporal(d, get_ctl_left_son(d, f)) && no_temporal(d, get_ctl_right_son(d, f));
    default :
      fprintf(stderr, "%s:%d:%s: fatal error\n", __FILE__, __LINE__, __FUNCTION__);
      abort();
  }
}

/* return 1 if is G xx with xx containing no temporal */
int isG(ctl_data *d, ctl_formula f)
{
  if (get_ctl_type(d, f) != ALWAYS)
    return 0;

  return no_temporal(d, get_ctl_son(d, f));
}

/* return 1 if is F xx with xx containing no temporal */
int isF(ctl_data *d, ctl_formula f)
{
  if (get_ctl_type(d, f) != EVENTUALLY)
    return 0;

  return no_temporal(d, get_ctl_son(d, f));
}

transition_tree onlyG(checker *c, int k)
{
  /* the translation for Gx =
   * Mk . Lk . sonG(0) . sonG(1) . ... . sonG(k)
   */
  transition_tree t, t1;
  int j;
  ctl_formula f1 = get_ctl_son(&c->ctl_data, c->formula);

  t = Mk(&c->trans_data, &c->variables, c->f_trans, c->init_states_trans, k);

  t1 = Lk(&c->trans_data, &c->variables, c->f_trans, k);

  t = new_transition_tree_op(&c->trans_data, AND, t, t1);

  t1 = translate_ltl(&c->trans_data, &c->variables, &c->ctl_data, f1, k, 0);
  for (j=1; j<=k; j++)
    t1 = new_transition_tree_op(&c->trans_data, AND, t1,
		translate_ltl(&c->trans_data, &c->variables, &c->ctl_data, f1, k, j));

  return new_transition_tree_op(&c->trans_data, AND, t, t1);
}

transition_tree onlyF(checker *c, int k)
{
  /* the translation of Fx =
   * Mk . ((sonF(0) + sonF(1) + ... + sonF(k))
   */
  transition_tree t, t1;
  int j;
  ctl_formula f1 = get_ctl_son(&c->ctl_data, c->formula);

  t = Mk(&c->trans_data, &c->variables, c->f_trans, c->init_states_trans, k);

  t1 = translate_ltl(&c->trans_data, &c->variables, &c->ctl_data, f1, k, 0);
  for (j=1; j<=k; j++)
    t1 = new_transition_tree_op(&c->trans_data, OR, t1,
		translate_ltl(&c->trans_data, &c->variables, &c->ctl_data, f1, k, j));

  return new_transition_tree_op(&c->trans_data, AND, t, t1);
}
