/*
 * A CTL parser.
 * The ctl tree definition.
 *
 *----
 * To my angel,
 * december 2001.
 *----
 * This file is in the public domain.
 */

#ifndef _CTL_H_
#define _CTL_H_

#include "vars.h"
#include "hash.h"
#include "bdd.h"

/* op values */
#define ATOM            0
#define ALL_PATH        1
#define ONE_PATH        2
#define ALWAYS          3
#define EVENTUALLY      4
#define UNTIL           5
#define NEXT            6
#define AND             7
#define OR              8
#define XOR		9
#define NOT             10
#define IMPLIES         11
/* those one are only present because of the parser */
#if 0
#define PAR_OPEN        12
#define PAR_CLOSE       13
#endif
#define CTL_TRUE	14
#define CTL_FALSE	15

typedef unsigned int ctl_formula;

#include <limits.h>

#define CTL_NO_FORMULA	UINT_MAX

typedef struct {
  int op;
  bdd *bdd;
  int nb_refs;
  union {
    variable variable;
    struct {
      ctl_formula s;
    } op_one;
    struct {
      ctl_formula l;
      ctl_formula r;
    } op_two;
  } val;
} _ctl_formula;

typedef struct {
  _ctl_formula *tab;
  int size;
  int maxsize;
  hash_table ht;
} ctl_data;

ctl_formula new_ctl_formula_var(ctl_data *, variable);
ctl_formula new_ctl_formula_op(ctl_data *, int, ctl_formula, ctl_formula);
ctl_formula new_ctl_formula_true(ctl_data *);
ctl_formula new_ctl_formula_false(ctl_data *);
int init_ctl(ctl_data *);
void free_ctl(ctl_data *);

void free_ctl_formula(ctl_data *, ctl_formula);
ctl_formula new_ctl_formula(ctl_data *);

/* when this function returns, the result is sure to be ok, otherwise it wouldn't have returned, so no check needed */
ctl_formula normalize_ctl_formula(ctl_data *, ctl_formula);

#define get_ctl_type(d, t) ((d)->tab[t].op)
#define set_ctl_type(d, t, v) (d)->tab[t].op = v
#define get_ctl_variable(d, t) ((d)->tab[t].val.variable)
#define set_ctl_variable(d, t, v) (d)->tab[t].val.variable = v
#define get_ctl_left_son(d, t) ((d)->tab[t].val.op_two.l)
#define set_ctl_left_son(d, t, v) (d)->tab[t].val.op_two.l = v
#define get_ctl_right_son(d, t) ((d)->tab[t].val.op_two.r)
#define set_ctl_right_son(d, t, v) (d)->tab[t].val.op_two.r = v
#define get_ctl_son(d, t) ((d)->tab[t].val.op_one.s)
#define set_ctl_son(d, t, v) (d)->tab[t].val.op_one.s = v

void print_formula(ctl_data *, variables *, ctl_formula);

#include <stdio.h>
void ctl_restart(FILE *);
int ctl_parse(void);
void ctl_lexer_close(void);
extern unsigned long ctl_linenumber;
int ctl_lex(void);
void ctl_error(const char *, ...);

#endif /* _CTL_H_ */
