%{

#include "checker.h"
#include "ctl_parser.h"
#include "verbose.h"

unsigned long ctl_linenumber;

/* we are in c99 mode, where fileno does not exist, so we declare it here */
#include <stdio.h>
int fileno(FILE *);

%}

%option nounput

space	[ \f\r\t\v]
letter	[a-zA-Z]
digit	[0-9]
symblet	[a-zA-Z_]
forall	"A"
forone	"E"
always	"G"
later	"F"
next	"X"
until	"U"
and	[&.]
or	[+|]
xor	[\^]
not	[!\-]
implies	"=>"

%x formula_type

%%

# {
  register int c;
  c = input();
  if (c == '!')
    BEGIN(formula_type);
  else {
    while (c != '\n' && c != EOF)
      c = input();
    if (c == '\n')
      ctl_linenumber++;
  }
}

"(" return '(';
")" return ')';
\; return ';';
{space} /* do nothing */
\n ctl_linenumber++;
{forall} return P_FORALL;
{forone} return P_FORONE;
{always} return P_ALWAYS;
{later} return P_LATER;
{next} return P_NEXT;
{until} return P_UNTIL;
{and} return P_AND;
{or} return P_OR;
{xor} return P_XOR;
{not} return P_NOT;
{implies} return P_IMPLIES;
True	return P_TRUE;
False	return P_FALSE;
[a-z_][A-Za-z_0-9]* ctl_lval.symbol = yytext; return P_SYMBOL;
. return P_ERROR;

<formula_type>ltl {
  print("ctl_parser", stdout, "An ltl formula is gonna be read\n");
  BEGIN(0);
  return P_TYPE_LTL;
}

<formula_type>\n ctl_linenumber++; BEGIN(0);
<formula_type>. {
  register int c;
  while ((c = input()) != '\n' && c != EOF)
    /* */;
  if (c == '\n')
    ctl_linenumber++;
  BEGIN(0);
}

%%

int yywrap(void)
{
  return 1;
}

void ctl_lexer_close(void)
{
  yy_delete_buffer(YY_CURRENT_BUFFER);
}
