%{
/* declarations C */
#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <stdarg.h>

#include "ctl.h"
#include "vars.h"
#include "checker.h"
#include "mem.h"
#include "verbose.h"

extern checker the_checker;

%}

/* definitions yacc */

%token <symbol> P_SYMBOL
%token P_FORALL P_FORONE P_ALWAYS P_LATER P_NEXT P_UNTIL P_AND P_OR P_XOR P_NOT P_IMPLIES P_TRUE P_FALSE
%token P_ERROR
%token P_TYPE_LTL

%type <formula> atom state_formula
%type <formula> state_atom_formula
%type <formula> state_formula_3
%type <formula> state_formula_4
%type <formula> state_formula_5
%type <formula> state_formula_6
%type <formula> ltl_formula
%type <formula> ltl_formula_1
%type <formula> ltl_formula_2
%type <formula> ltl_formula_3
%type <formula> ltl_formula_4
%type <formula> ltl_formula_5

%start ctl_formula

%union {
  char *symbol;
  ctl_formula formula;
}

%%

/* regles de production */

ctl_formula :
	  P_TYPE_LTL
	  ltl_formula
		{
		  the_checker.formula = $2;
		  the_checker.formula_type = TYPE_LTL;

		  print("ctl_parser", stdout, "here is the ltl formula before transformation :\n");
		  if (verbose_on("ctl_parser"))
		    print_formula(&the_checker.ctl_data, &the_checker.variables, the_checker.formula);
		  print("ctl_parser", stdout, "\n");
		}
	| state_formula
		{
		  the_checker.formula = $1;
		  the_checker.formula_type = TYPE_CTL;

		  print("ctl_parser", stdout, "here is the ctl formula before transformation :\n");
		  if (verbose_on("ctl_parser"))
		    print_formula(&the_checker.ctl_data, &the_checker.variables, the_checker.formula);
		  print("ctl_parser", stdout, "\n");
		}
	| state_formula ';'
	  fairness
		{
		  the_checker.formula = $1;
		  the_checker.formula_type = TYPE_CTL;

		  print("ctl_parser", stdout, "here is the ctl formula before transformation :\n");
		  if (verbose_on("ctl_parser"))
		    print_formula(&the_checker.ctl_data, &the_checker.variables, the_checker.formula);
		  print("ctl_parser", stdout, "\n");
		}
	;

fairness :
	  /* nothing */
	| constraint_list
	;

constraint_list :
	  state_formula ';'
		{
		  the_checker.nbmax_fairness = 32;
		  the_checker.fairness = getmem(32 * sizeof(ctl_formula));
		  the_checker.fairness[0] = $1;
		  the_checker.nb_fairness = 1;
		}
	| constraint_list state_formula ';'
		{
		  if (the_checker.nbmax_fairness == the_checker.nb_fairness) {
		    the_checker.nbmax_fairness += 32;
		    the_checker.fairness = regetmem(the_checker.fairness, the_checker.nbmax_fairness * sizeof(ctl_formula));
		  }
		  the_checker.fairness[the_checker.nb_fairness++] = $2;
		}
	;

state_atom_formula :
	  atom
		{
		  $$ = $1;
		}
	| P_NOT state_atom_formula
		{
		  $$ = new_ctl_formula_op(&the_checker.ctl_data, NOT, $2, CTL_NO_FORMULA);
		}
	| '(' state_formula ')'
		{
		  $$ = $2;
		}
	| P_FORALL P_ALWAYS state_atom_formula
		{
		  $$ = new_ctl_formula_op(&the_checker.ctl_data, ALL_PATH,
			 new_ctl_formula_op(&the_checker.ctl_data, ALWAYS, $3, CTL_NO_FORMULA), CTL_NO_FORMULA);
		}
	| P_FORONE P_ALWAYS state_atom_formula
		{
		  $$ = new_ctl_formula_op(&the_checker.ctl_data, ONE_PATH,
			 new_ctl_formula_op(&the_checker.ctl_data, ALWAYS, $3, CTL_NO_FORMULA), CTL_NO_FORMULA);
		}
	| P_FORALL P_LATER state_atom_formula
		{
		  $$ = new_ctl_formula_op(&the_checker.ctl_data, ALL_PATH,
			 new_ctl_formula_op(&the_checker.ctl_data, EVENTUALLY, $3, CTL_NO_FORMULA), CTL_NO_FORMULA);
		}
	| P_FORONE P_LATER state_atom_formula
		{
		  $$ = new_ctl_formula_op(&the_checker.ctl_data, ONE_PATH,
			 new_ctl_formula_op(&the_checker.ctl_data, EVENTUALLY, $3, CTL_NO_FORMULA), CTL_NO_FORMULA);
		}
	| P_FORALL P_NEXT state_atom_formula
		{
		  $$ = new_ctl_formula_op(&the_checker.ctl_data, ALL_PATH,
			 new_ctl_formula_op(&the_checker.ctl_data, NEXT, $3, CTL_NO_FORMULA), CTL_NO_FORMULA);
		}
	| P_FORONE P_NEXT state_atom_formula
		{
		  $$ = new_ctl_formula_op(&the_checker.ctl_data, ONE_PATH,
			 new_ctl_formula_op(&the_checker.ctl_data, NEXT, $3, CTL_NO_FORMULA), CTL_NO_FORMULA);
		}
	;

state_formula_3 :
	  state_atom_formula
		{
		  $$ = $1;
		}
	| state_formula_3 P_AND state_atom_formula
		{
		  $$ = new_ctl_formula_op(&the_checker.ctl_data, AND, $1, $3);
		}
	;

state_formula_4 :
	  state_formula_3
		{
		  $$ = $1;
		}
	| state_formula_4 P_OR state_formula_3
		{
		  $$ = new_ctl_formula_op(&the_checker.ctl_data, OR, $1, $3);
		}
	| state_formula_4 P_XOR state_formula_3
		{
		  $$ = new_ctl_formula_op(&the_checker.ctl_data, XOR, $1, $3);
		}
	;

state_formula_5 :
	  state_formula_4
		{
		  $$ = $1;
		}
	| state_formula_4 P_IMPLIES state_formula_5
		{
		  $$ = new_ctl_formula_op(&the_checker.ctl_data, IMPLIES, $1, $3);
		}
	;

state_formula_6 :
	  state_formula_5
		{
		  $$ = $1;
		}
	| P_FORALL '(' state_formula_6 P_UNTIL state_formula_6 ')'
		{
		  $$ = new_ctl_formula_op(&the_checker.ctl_data, ALL_PATH,
			 new_ctl_formula_op(&the_checker.ctl_data, UNTIL, $3, $5), CTL_NO_FORMULA);
		}
	| P_FORONE '(' state_formula_6 P_UNTIL state_formula_6 ')'
		{
		  $$ = new_ctl_formula_op(&the_checker.ctl_data, ONE_PATH,
			 new_ctl_formula_op(&the_checker.ctl_data, UNTIL, $3, $5), CTL_NO_FORMULA);
		}
	;

state_formula :
	  state_formula_6
		{
		  $$ = $1;
		}
	;

atom :
	  P_SYMBOL
		{
		  variable v = get_variable(&the_checker.variables, $1);
		  if (v == NO_VARIABLE) {
		    yyerror("variable undefined %s", $1);
		    YYERROR;
		  }
		  $$ = new_ctl_formula_var(&the_checker.ctl_data, v);
		}
	| P_TRUE
		{
		  $$ = new_ctl_formula_true(&the_checker.ctl_data);
		}
	| P_FALSE
		{
		  $$ = new_ctl_formula_false(&the_checker.ctl_data);
		}
	;

ltl_formula :
	  ltl_formula_1
		{
		  $$ = $1;
		}
	;

ltl_formula_1 :
	  ltl_formula_2
		{
		  $$ = $1;
		}
	| ltl_formula_2 P_IMPLIES ltl_formula_1
		{
		  $$ = new_ctl_formula_op(&the_checker.ctl_data, IMPLIES, $1, $3);
		}
	;

ltl_formula_2 :
	  ltl_formula_3
		{
		  $$ = $1;
		}
	| ltl_formula_3 P_OR ltl_formula_2
		{
		  $$ = new_ctl_formula_op(&the_checker.ctl_data, OR, $1, $3);
		}
	;

ltl_formula_3 :
	  ltl_formula_4
		{
		  $$ = $1;
		}
	| ltl_formula_4 P_AND ltl_formula_3
		{
		  $$ = new_ctl_formula_op(&the_checker.ctl_data, AND, $1, $3);
		}
	;

ltl_formula_4 :
	  ltl_formula_5
		{
		  $$ = $1;
		}
	| ltl_formula_5 P_UNTIL ltl_formula_4
		{
		  $$ = new_ctl_formula_op(&the_checker.ctl_data, UNTIL, $1, $3);
		}
	;

ltl_formula_5 :
	  P_LATER ltl_formula_5
		{
		  $$ = new_ctl_formula_op(&the_checker.ctl_data, EVENTUALLY, $2, CTL_NO_FORMULA);
		}
	| P_ALWAYS ltl_formula_5
		{
		  $$ = new_ctl_formula_op(&the_checker.ctl_data, ALWAYS, $2, CTL_NO_FORMULA);
		}
	| P_NEXT ltl_formula_5
		{
		  $$ = new_ctl_formula_op(&the_checker.ctl_data, NEXT, $2, CTL_NO_FORMULA);
		}
	| P_NOT ltl_formula_5
		{
		  $$ = new_ctl_formula_op(&the_checker.ctl_data, NOT, $2, CTL_NO_FORMULA);
		}
	| atom
		{
		  $$ = $1;
		}
	| '(' ltl_formula ')'
		{
		  $$ = $2;
		}
	;

%%

/* fonctions C */

void ctl_error(const char *s, ...)
{
  va_list a;
  va_start(a, s);
  vfprintf(stderr, s, a);
  fprintf(stderr, " line %ld\n", ctl_linenumber);
  va_end(a);
}
