%{
/* section declarations, inclues dans le resultat */

/* we must include this before fsm_parser.h because this one does define any structure, even the one it uses in %union, which we can understand */
#include "bdd.h"
#include "ctl.h"
#include "vars.h"
#include "checker.h"
#include "trans.h"
#include "verbose.h"

#include "fsm_parser.h"

unsigned long fsm_linenumber;

/* we are in c99 mode, where fileno does not exist, so we declare it here */
#include <stdio.h>
int fileno(FILE *);

%}

%option nounput

/* section macro "x = m ;" */
space	[ \f\r\t\v]
letter	[a-zA-Z]
digit	[0-9]
symblet	[a-zA-Z_]
forall	"A"
forone	"E"
always	"G"
later	"F"
next	"X"
until	"U"
and	[&.]
or	[+|]
xor	[\^]
not	[!\-]
implies	"=>"|"->"
equal "="

/* section regles */
/*
 *  $ fin de ligne
 *  ^ debut de ligne sauf entre crochet ou c'est le complementaire
 *  . tout sauf \n
 *  - entre crochet definit un intervalle
 * <> definit une condition de debut
 * [] definit un sous-ens de cars
 *  ? la regexp qui precede est facultative
 *  * repetition de exp pred 0 a n fois
 *  +                        1 a n fois
 *  | definit des exp alternees
 * () groupe des exp
 *  / definit une suite d'exp
 * {} definit une repetition explicite ou sert de regle de substitution pour les macros lex
 * \ inhibe la signification du carac suivant
 * " definit une suite de cars
 */

/* eat up comments (in a fast and pretty way) */

%x formula_type

%%

# {
  register int c;
  c = input();
  if (c == '!') {
    BEGIN(formula_type);
  } else {
    while (c != '\n' && c != EOF)
      c = input();
    if (c == '\n')
      fsm_linenumber++;
  }
}

"(" return '(';
")" return ')';
{space} /* do nothing */
\n fsm_linenumber++;
{and} return P_AND;
{or} return P_OR;
{not} return P_NOT;
{xor} return P_XOR;
{implies} return P_IMPLIES;
{equal} return P_EQUAL;
\; return P_SEPAR;
True|true|T return P_TRUE;
False|false|B return P_FALSE;
[a-z_][A-Za-z_0-9]* fsm_lval.symbol = yytext; return P_SYMBOL;
. return P_ERROR;

<formula_type>postfix {
  print("fsm_parser", stdout, "A postfix fsm is gonna be read\n");
  BEGIN(0);
  return P_TYPE_POSTFIX;
}
<formula_type>infix {
  print("fsm_parser", stdout, "An infix fsm is gonna be read\n");
  BEGIN(0);
  return P_TYPE_INFIX;
}
<formula_type>\n fsm_linenumber++; BEGIN(0);
<formula_type>. {
  register int c;
  while ((c = input()) != '\n' && c != EOF)
    /* */;
  if (c == '\n')
    fsm_linenumber++;
  BEGIN(0);
}

%%

/* fonctions utilisateurs (soit main en general) */

int yywrap(void)
{
  return 1;
}

void fsm_lexer_close(void)
{
  yy_delete_buffer(YY_CURRENT_BUFFER);
}

