%{
/* declarations C */
#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <stdarg.h>

#include "bdd.h"
#include "ctl.h"
#include "vars.h"
#include "checker.h"
#include "trans.h"
#include "verbose.h"

extern checker the_checker;

%}

/* definitions yacc */

%token <symbol> P_SYMBOL
%token P_AND P_XOR P_OR P_NOT P_IMPLIES P_TRUE P_FALSE P_EQUAL
%token P_ERROR P_SEPAR

%token P_TYPE_POSTFIX P_TYPE_INFIX

%type <var> existing_variable
%type <var_set> variables_list
%type <trans_tree> transition_function atom init_state function
%type <trans_tree> transition_function_infix tf_inf0 tf_inf1 tf_inf2 tf_inf3 tf_inf4 init_state_infix function_infix

%union {
  variable var;
  char *symbol;
  bdd_variables_set *var_set;
  transition_tree trans_tree;
}

%start fsm_file

%%

/* regles de production */
fsm_file :
	  fsm_postfix  /* postfix by default, for compatibility reasons */
	| P_TYPE_POSTFIX fsm_postfix
	| P_TYPE_INFIX fsm_infix
	;

fsm_postfix :
	  variables_declaration P_SEPAR
	  primary_variables P_SEPAR
	  secondary_variables P_SEPAR
	  input_variables P_SEPAR
	  output_variables P_SEPAR
          init_state
	  function
	  output_update
		{
		  print("fsm_parser", stdout, "fsm_parser.y:fsm_postfix: all parsed ok\n");
		  /*
		  the_checker.vars = $3;
		  the_checker.varsnext = $5;
		  the_checker.varsinput = $7;
		  */
		  if (the_checker.we_do != WE_DO_BMC) {
		    bdd_update_variables_set(&the_checker.bdd_data, the_checker.vars);
		    bdd_update_variables_set(&the_checker.bdd_data, the_checker.varsnext);
		    bdd_update_variables_set(&the_checker.bdd_data, the_checker.varsinput);
		    the_checker.init_states = transition_tree_to_bdd(&the_checker.bdd_data, &the_checker.trans_data, &the_checker.variables, $11);
		    the_checker.f = transition_tree_to_bdd(&the_checker.bdd_data, &the_checker.trans_data, &the_checker.variables, $12);
		    print("fsm_parser", stdout, "fsm_parser.y:fsm_postfix: bdd have been builded\n");
		  }
		  the_checker.init_states_trans = $11;
		  the_checker.f_trans = $12;
		}
	;

output_update :
	  /* nothing, for compatibility */
	| output_update existing_variable function
		{
		  /* here, one should free the bdd associated to $2 */
		  if (the_checker.we_do != WE_DO_BMC) {
		    set_variable_assoc(&the_checker.variables, $2,
		      transition_tree_to_bdd(&the_checker.bdd_data, &the_checker.trans_data, &the_checker.variables, $3));
		    free_transition_tree(&the_checker.trans_data, $3);
		  }
		}
	;

existing_variable :
	  P_SYMBOL
		{
		  $$ = get_variable(&the_checker.variables, $1);
		  if ($$ == NO_VARIABLE) {
		    yyerror("undefined variable %s\n", $1);
		    YYERROR;
		  }
		}
	;

variables_declaration :
	  P_SYMBOL
		{
		  variable v;
		  if ((v=new_variable(&the_checker.variables, $1))==-1) {
		    yyerror("variable already defined %s", $1);
		    YYERROR;
		  }
		  if (the_checker.we_do != WE_DO_BMC)
		    set_variable_assoc(&the_checker.variables, v, bdd_new_variable(&the_checker.bdd_data));
		}
	| variables_declaration P_SYMBOL
		{
		  variable v;
		  if ((v=new_variable(&the_checker.variables, $2))==-1) {
		    yyerror("variable already defined %s", $2);
		    YYERROR;
		  }
		  if (the_checker.we_do != WE_DO_BMC)
		    set_variable_assoc(&the_checker.variables, v, bdd_new_variable(&the_checker.bdd_data));
		}
	;

primary_variables :
	  variables_list
		{
		  the_checker.vars = $1;
		  set_variables_set_type(&the_checker.variables, $1, VARIABLE_TYPE_CURRENT);
		}
	;

secondary_variables :
	  variables_list
		{
		  the_checker.varsnext = $1;
		  set_variables_set_type(&the_checker.variables, $1, VARIABLE_TYPE_NEXT);
		}
	;

input_variables :
	  variables_list
		{
		  the_checker.varsinput = $1;
		  set_variables_set_type(&the_checker.variables, $1, VARIABLE_TYPE_INPUT);
		}
	;

output_variables :
	  /* nothing */
	  /* to remain compatible with old version */
	  {}
	| variables_list
		{
		  the_checker.varsoutput = $1;
		  set_variables_set_type(&the_checker.variables, $1, VARIABLE_TYPE_OUTPUT);
		}
	;

variables_list :
	  P_SYMBOL
		{
		  bdd_variables_set *s = new_bdd_variables_set();
		  variable v = get_variable(&the_checker.variables, $1);
		  bdd *b = 0;

		  if (the_checker.we_do != WE_DO_BMC) {
		    b = (bdd *)get_variable_assoc(&the_checker.variables, v);
		    if (!b) {
		      free_bdd_variables_set(&the_checker.bdd_data, s);
		      yyerror("undefined variable %s", $1);
		      YYERROR;
		    }
		  }

		  bdd_variables_set_add(s, v, b);
		  $$ = s;
		}
	| variables_list P_SYMBOL
		{
		  variable v = get_variable(&the_checker.variables, $2);
		  bdd *b = 0;

		  if (the_checker.we_do != WE_DO_BMC) {
		    b = (bdd *)get_variable_assoc(&the_checker.variables, v);
		    if (!b) {
		      free_bdd_variables_set(&the_checker.bdd_data, $1);
		      yyerror("undefined variable %s", $2);
		      YYERROR;
		    }
		  }

		  bdd_variables_set_add($1, v, b);
		  $$ = $1;
		}
	;

init_state :
	  transition_function P_SEPAR
		{
		  print("fsm_parser", stdout, "here comes the init state :\n");
		  if (verbose_on("fsm_parser"))
		    print_tree(&the_checker.trans_data, &the_checker.variables, $1);
		  print("fsm_parser", stdout, "\n");
		  $$ = $1;
		}
	;

function :
	  transition_function P_SEPAR
		{
		  print("fsm_parser", stdout, "here comes the transition function :\n");
		  if (verbose_on("fsm_parser"))
		    print_tree(&the_checker.trans_data, &the_checker.variables, $1);
		  print("fsm_parser", stdout, "\n");
		  $$ = $1;
		}
	;

transition_function :
	  atom
		{
		  $$ = $1;
		}
	| transition_function P_NOT
		{
		  $$ = new_transition_tree_op(&the_checker.trans_data, NOT, $1, TRANS_UNDEF);
		}
	| transition_function transition_function P_AND
		{
		  $$ = new_transition_tree_op(&the_checker.trans_data, AND, $1, $2);
		}
	| transition_function transition_function P_XOR
		{
		  $$ = new_transition_tree_op(&the_checker.trans_data, XOR, $1, $2);
		}
	| transition_function transition_function P_OR
		{
		  $$ = new_transition_tree_op(&the_checker.trans_data, OR, $1, $2);
		}
	| transition_function transition_function P_IMPLIES
		{
		  $$ = new_transition_tree_op(&the_checker.trans_data, IMPLIES, $1, $2);
		}
	| transition_function transition_function P_EQUAL
		{
		  $$ = new_transition_tree_op(&the_checker.trans_data, EQUAL, $1, $2);
		}
	;

atom :
	  P_SYMBOL
		{
		  variable v = get_variable(&the_checker.variables, $1);
		  if (v==-1) {
		    yyerror("unknown variable %s", $1);
		    YYERROR;
		  }
		  $$ = new_transition_tree_var(&the_checker.trans_data, v);
		}
	| P_TRUE
		{
		  $$ = new_transition_tree_true(&the_checker.trans_data);
		}
	| P_FALSE
		{
		  $$ = new_transition_tree_false(&the_checker.trans_data);
		}
	;

fsm_infix :
	  variables_declaration P_SEPAR
	  primary_variables P_SEPAR
	  secondary_variables P_SEPAR
	  input_variables P_SEPAR
	  output_variables P_SEPAR
          init_state_infix
	  function_infix
	  output_update_infix
		{
		  if (the_checker.we_do != WE_DO_BMC) {
		    bdd_update_variables_set(&the_checker.bdd_data, the_checker.vars);
		    bdd_update_variables_set(&the_checker.bdd_data, the_checker.varsnext);
		    bdd_update_variables_set(&the_checker.bdd_data, the_checker.varsinput);
		    the_checker.init_states = transition_tree_to_bdd(&the_checker.bdd_data, &the_checker.trans_data, &the_checker.variables, $11);
		    the_checker.f = transition_tree_to_bdd(&the_checker.bdd_data, &the_checker.trans_data, &the_checker.variables, $12);
		  }
		  the_checker.init_states_trans = $11;
		  the_checker.f_trans = $12;
		}
	;

output_update_infix :
	  /* nothing, for compatibility */
	| output_update_infix existing_variable function_infix
		{
		  if (the_checker.we_do != WE_DO_BMC) {
		    /* here, one should free the bdd associated to $2 */
		    set_variable_assoc(&the_checker.variables, $2,
		      transition_tree_to_bdd(&the_checker.bdd_data, &the_checker.trans_data, &the_checker.variables, $3));
		    free_transition_tree(&the_checker.trans_data, $3);
		  }
		}
	;

init_state_infix :
	  transition_function_infix P_SEPAR
		{
		  print("fsm_parser", stdout, "here comes the init state :\n");
		  if (verbose_on("fsm_parser"))
		    print_tree(&the_checker.trans_data, &the_checker.variables, $1);
		  print("fsm_parser", stdout, "\n");
		  $$ = $1;
		}
	;

function_infix :
	  transition_function_infix P_SEPAR
		{
		  print("fsm_parser", stdout, "here comes the transition function :\n");
		  if (verbose_on("fsm_parser"))
		    print_tree(&the_checker.trans_data, &the_checker.variables, $1);
		  print("fsm_parser", stdout, "\n");
		  $$ = $1;
		}

transition_function_infix :
	  tf_inf4 {$$=$1;}
	;

tf_inf0 :
	  atom {$$=$1;}
	| P_NOT tf_inf0
		{
		  $$ = new_transition_tree_op(&the_checker.trans_data, NOT, $2, TRANS_UNDEF);
		}
	| '(' transition_function_infix ')' {$$=$2;}
	;

tf_inf1 :
	  tf_inf0 {$$=$1;}
	| tf_inf1 P_AND tf_inf0
		{
		  $$ = new_transition_tree_op(&the_checker.trans_data, AND, $1, $3);
		}
	;

tf_inf2 :
	  tf_inf1 {$$=$1;}
	| tf_inf2 P_OR tf_inf1
		{
		  $$ = new_transition_tree_op(&the_checker.trans_data, OR, $1, $3);
		}
	| tf_inf2 P_XOR tf_inf1
		{
		  $$ = new_transition_tree_op(&the_checker.trans_data, XOR, $1, $3);
		}
	;

tf_inf3 :
	  tf_inf2 {$$=$1;}
	| tf_inf3 P_IMPLIES tf_inf2
		{
		  $$ = new_transition_tree_op(&the_checker.trans_data, IMPLIES, $1, $3);
		}
	;

tf_inf4 :
	  tf_inf3
	| tf_inf4 P_EQUAL tf_inf3
		{
		  $$ = new_transition_tree_op(&the_checker.trans_data, EQUAL, $1, $3);
		}
	;

%%

/* fonctions C */

#if 0
int main(void)
{
  yyparse();
  return 0;
}
#endif

extern unsigned long fsm_linenumber;

void fsm_error(const char *s, ...)
{
  va_list a;
  fprintf(stderr, "%s:%ld: ", the_checker.filename, fsm_linenumber);
  va_start(a, s);
  vfprintf(stderr, s, a);
  va_end(a);
  fprintf(stderr, "\n");
}
