/*
 * A hash table - more or less generic
 *---
 * to my angel
 * public domain
 * march 2002
 */

#include "hash2.h"

#include <string.h>
#include <stdio.h>
#include <stdlib.h>

#include "mem.h"
#include "verbose.h"

static inline int bit_on(unsigned char *c, unsigned long i)
{
  register unsigned long ii = i >> 3;
  register int iii = i & 7;

  return c[ii] & (1<<iii);
}

static inline void set_bit(unsigned char *c, unsigned long i)
{
  register unsigned long ii = i >> 3;
  register int iii = i & 7;

  c[ii] |= (1 << iii);
}

/* a forward declaration */
void init_expand(void);

/* h->flag must be used to test if all is ok */
void init_hash2(hash2_table *h, int init_size)
{
  memset(h, 0, sizeof(hash2_table));

  h->data = getmem(sizeof(elem) * init_size);
  h->data_busy = getmem(sizeof(unsigned char) * init_size / 8);
  h->maxsize = init_size;
  memset(h->data_busy, 0, sizeof(unsigned char) * init_size / 8);

  init_expand();

  h->flag = HASH2_NO_ERROR;
}

void free_hash2(hash2_table *h)
{
  print("hashstat", stdout, "hash stat : %ld collisions, final hash size %ld, nb items %ld\n", h->hash_collision, h->maxsize, h->size);

  if (h->maxsize) {
    freemem(h->data);
    freemem(h->data_busy);
  }

  memset(h, 0, sizeof(hash2_table));

  h->flag = HASH2_NO_ERROR;
}

/* when collision, we call hash_next to get the next hash value */
static unsigned long hash2_next(hash2_table *h, unsigned long hash)
{
  /* some stats */
  h->hash_collision++;

  return hash+h->hash_random;
}

static unsigned short expanded1[256];
static unsigned short expanded2[256];
static char inited_expand = 0;

void init_expand(void)
{
  int i;

  if (inited_expand)
    return;

  inited_expand = 1;

  for (i=0; i<256; i++) {
    unsigned short a;
    unsigned char b;
    a = 0;
    b = i;
    a |= (b & 1);
    a |= (b & 2)<<1;
    a |= (b & 4)<<2;
    a |= (b & 8)<<3;
    a |= (b & 16)<<4;
    a |= (b & 32)<<5;
    a |= (b & 64)<<6;
    a |= (b & 128)<<7;
    expanded1[i] = a;
    expanded2[i] = a<<1;
  }
}

static unsigned long hash2(hash2_table *h, unsigned long op, unsigned long l, unsigned long r)
{
#if 1
  /* first hash, very simple, very basic, but efficient ?
   * hash stat : 1081518 collisions, final hash size 524288, nb items 248188 */
  h->hash_random = (l+r+op) | 1;

  return l+r+op+((l+r+op)<<16);
#elif 0
  /* second hash : a little more elaborated
   * hash stat : 2502609 collisions, final hash size 524288, nb items 231818 */
  unsigned long ret =  ((expanded1[l&255] | (expanded1[(l>>8)&255]<<16) |
         expanded2[r&255] | (expanded2[(r>>8)&255]<<16)) << 4) | (op & 15);

  h->hash_random = ((l+r+op) + ((l+r+op)<<16)) | 1;

  return ret;

#else
#endif
}

static void rehash2(hash2_table *h)
{
  unsigned long i;
  unsigned long hash;
  unsigned long index;
  unsigned long oldsize = h->maxsize;
  elem *old_data = h->data;
  elem *od2 = old_data;
  unsigned char *old_busy = h->data_busy;
  unsigned char *ob2 = old_busy;

  print("hash", stdout, "rehash2 called old_size = %lu\n", oldsize);

  h->maxsize *= 2;
  if (h->maxsize == 0x80000000) {
    fprintf(stderr, "%s:%d:%s: fatal error, out of memory\n", __FILE__, __LINE__, __FUNCTION__);
    abort();
  }

  h->data = getmem(h->maxsize * sizeof(elem));
  h->data_busy = getmem(h->maxsize * sizeof(unsigned char) / 8);

  memset(h->data_busy, 0, sizeof(unsigned char) * h->maxsize / 8);

  for (i=0; i<oldsize; i++, old_data ++) {
    if (!bit_on(old_busy, i))
      continue;

    hash = hash2(h, old_data->op, old_data->l, old_data->r);
    index = hash & (h->maxsize - 1);

    /* collision handling */
    while (bit_on(h->data_busy, index)) {
      hash = hash2_next(h, hash);
      index = hash & (h->maxsize - 1);
    }

    set_bit(h->data_busy, index);
    memcpy(&h->data[index], old_data, sizeof(elem));
  }

  freemem(od2);
  freemem(ob2);
}

unsigned long hash2_get_new_slot(hash2_table *h, unsigned long op, unsigned long l, unsigned long r)
{
  unsigned long hash;
  unsigned long index;

  if (h->size * 4 >= h->maxsize * 3)
    rehash2(h);

  hash = hash2(h, op, l, r);
  index = hash & (h->maxsize - 1);

  print("hash", stdout, "new slot %lu %lu %lu (index = %lu)\n", op, l, r, index);

  /* collision detection */
  while (bit_on(h->data_busy, index)) {
    //if (h->compare(h->compare_pointer, d, hash_get_data(h, index)) == 1) {
    if (h->data[index].op == op && h->data[index].l == l && h->data[index].r == r) {
      h->flag = HASH2_ALREADY_PRESENT;
      return index;
    }
    hash = hash2_next(h, hash);
    index = hash & (h->maxsize - 1);
  }

  set_bit(h->data_busy, index);

  h->size++;
  h->flag = HASH2_NO_ERROR;
  return index;
}

void hash2_feed_slot(hash2_table *ht, unsigned long h, unsigned long t, unsigned long op, unsigned long l, unsigned long r)
{
  ht->data[h].tree = t;
  ht->data[h].op = op;
  ht->data[h].l = l;
  ht->data[h].r = r;

  ht->flag = HASH2_NO_ERROR;
}
