/********************************************************************/
/*                                                                  */
/* to_cnf2, the real one, creates variables on the fly, on per node */
/*                                                                  */
/********************************************************************/
static int nn;
static char *vars;
static long vars_size;


static void numberize(trans_data *d, transition_tree t)
{
  if (d->trans_tree[t].number)
    return;

  d->trans_tree[t].number = nn;
  nn++;

  switch(get_transition_tree_type(d, t)) {
    case VAR :
    case CTL_TRUE :
    case CTL_FALSE :
      return;
    case NOT :
      numberize(d, get_transition_tree_son(d, t));
      return;
    case AND :
    case OR :
    case XOR :
    case EQUAL :
    case IMPLIES :
      numberize(d, get_transition_tree_left_son(d, t));
      numberize(d, get_transition_tree_right_son(d, t));
      return;
    default :
      fprintf(stderr, "trans.c:%d:numberize: error : unhandled type %d\n", __LINE__, get_transition_tree_type(d, t));
      abort();
  }
}

/* return 0 if the clause is known to be true (v + !v is <-> to true) */
/* this should not happen */
static int mark_atomic_clause(char *tab, trans_data *d, transition_tree t)
{
  switch(get_transition_tree_type(d, t)) {
    case VAR :
      if (tab[get_transition_tree_variable(d, t)] == 2)
	return 0;
      tab[get_transition_tree_variable(d, t)] = 1;
      return 1;
    case NOT : {
      transition_tree n = get_transition_tree_son(d, t);
      if (get_transition_tree_type(d, n) != VAR) {
	fprintf(stderr, "%s:%d:%s: error : type should be VAR\n", __FILE__, __LINE__, __FUNCTION__);
	abort();
      }
      if (tab[get_transition_tree_variable(d, n)] == 1)
	return 0;
      tab[get_transition_tree_variable(d, n)] = 2;
      return 1;
    }
    case OR :
      if (mark_atomic_clause(tab, d, get_transition_tree_left_son(d, t)) == 0)
	return 0;
      if (mark_atomic_clause(tab, d, get_transition_tree_right_son(d, t)) == 0)
	return 0;
      return 1;
    default :
      fprintf(stderr, "%s:%d:%s: error : unhandled type %d\n", __FILE__, __LINE__, __FUNCTION__, get_transition_tree_type(d, t));
      abort();
  }
}

static void unmark(char *tab, trans_data *d, transition_tree t)
{
  switch(get_transition_tree_type(d, t)) {
    case VAR :
      tab[get_transition_tree_variable(d, t)] = 0;
      return;
    case NOT : {
      transition_tree n = get_transition_tree_son(d, t);
      if (get_transition_tree_type(d, n) != VAR) {
	fprintf(stderr, "%s:%d:%s: error : type should be VAR\n", __FILE__, __LINE__, __FUNCTION__);
	abort();
      }
      tab[get_transition_tree_variable(d, n)] = 0;
      return;
    }
    case OR :
      unmark(tab, d, get_transition_tree_left_son(d, t));
      unmark(tab, d, get_transition_tree_right_son(d, t));
      return;
    default :
      fprintf(stderr, "%s:%d:%s: error : unhandled type %d\n", __FILE__, __LINE__, __FUNCTION__, get_transition_tree_type(d, t));
      abort();
  }
}

static void add_atomic_clause(clauses *c, trans_data *d, transition_tree t)
{
  int i;
  clause *l;

  if (mark_atomic_clause(vars, d, t) == 0) {
    unmark(vars, d, t);
    return;
  }

  l = new_clause(c);

  /* bad loop, should mix up stuff with mark_atomic_clause or use a stack to remember the vars */
  for (i=1; i<nn; i++) {
    if (vars[i] == 1) add_atom(l, i);
    else if (vars[i] == 2) add_atom(l, -i);
  }

  unmark(vars, d, t);
}

static void add_all_atomic_clauses(clauses *c, trans_data *d, transition_tree t)
{
  switch(get_transition_tree_type(d, t)) {
    case AND :
      add_all_atomic_clauses(c, d, get_transition_tree_left_son(d, t));
      add_all_atomic_clauses(c, d, get_transition_tree_right_son(d, t));
      return;
    case OR :
      add_atomic_clause(c, d, t);
      return;
    default :
      fprintf(stderr, "%s:%d:%s: error : unhandled type %d\n", __FILE__, __LINE__, __FUNCTION__, get_transition_tree_type(d, t));
      abort();
  }
}

static void closify(trans_data *nd, clauses *cl, trans_data *d, transition_tree t)
{
#if 0
transition_tree t1, t2, t3;
t1 = new_transition_tree_var(d, 10);
t2 = new_transition_tree_var(d, 11);

d->trans_tree[t1].number = 10;
d->trans_tree[t2].number = 11;
t3 = new_transition_tree_op(d, IMPLIES, t1, t2);
d->trans_tree[t3].number = 12;
add_clause(nd, c, d, t3);
return;
#endif

#if 1
  if (d->trans_tree[t].added == 1)
    return;

  d->trans_tree[t].added = 1;

  switch(get_transition_tree_type(d, t)) {
    case VAR :
    case CTL_TRUE :
    case CTL_FALSE :
      return;
    case NOT : { /* a = !b == (a + b) . (!a + !b) */
      unsigned long a = d->trans_tree[t].number;
      unsigned long b = d->trans_tree[get_transition_tree_son(d, t)].number;
      clause *l;

      l = new_clause(cl);
      add_atom(l, a);
      add_atom(l, b);

      l = new_clause(cl);
      add_atom(l, -a);
      add_atom(l, -b);

      return closify(nd, cl, d, get_transition_tree_son(d, t));
    }

    case AND : { /* a = b . c == (a + !b + !c) . (!a + b) . (!a + c) */
      unsigned long a = d->trans_tree[t].number;
      unsigned long b = d->trans_tree[get_transition_tree_left_son(d, t)].number;
      unsigned long c = d->trans_tree[get_transition_tree_right_son(d, t)].number;
      clause *l;

      l= new_clause(cl);
      add_atom(l, a);
      add_atom(l, -b);
      add_atom(l, -c);

      l= new_clause(cl);
      add_atom(l, -a);
      add_atom(l, b);

      l= new_clause(cl);
      add_atom(l, -a);
      add_atom(l, c);

      closify(nd, cl, d, get_transition_tree_left_son(d, t));
      return closify(nd, cl, d, get_transition_tree_right_son(d, t));
    }

    case OR : { /* a = b + c == (!a + b + c) . (a + !b) . (a + !c) */
      unsigned long a = d->trans_tree[t].number;
      unsigned long b = d->trans_tree[get_transition_tree_left_son(d, t)].number;
      unsigned long c = d->trans_tree[get_transition_tree_right_son(d, t)].number;
      clause *l;

      l= new_clause(cl);
      add_atom(l, -a);
      add_atom(l, b);
      add_atom(l, c);

      l= new_clause(cl);
      add_atom(l, a);
      add_atom(l, -b);

      l= new_clause(cl);
      add_atom(l, a);
      add_atom(l, -c);

      closify(nd, cl, d, get_transition_tree_left_son(d, t));
      return closify(nd, cl, d, get_transition_tree_right_son(d, t));
    }

    case XOR : { /* a = b xor c == (a + !b + c) . (a + b + !c) . (!a + b + c) . (!a + !b + !c) */
      unsigned long a = d->trans_tree[t].number;
      unsigned long b = d->trans_tree[get_transition_tree_left_son(d, t)].number;
      unsigned long c = d->trans_tree[get_transition_tree_right_son(d, t)].number;
      clause *l;

      l= new_clause(cl);
      add_atom(l, a);
      add_atom(l, -b);
      add_atom(l, c);

      l= new_clause(cl);
      add_atom(l, a);
      add_atom(l, b);
      add_atom(l, -c);

      l= new_clause(cl);
      add_atom(l, -a);
      add_atom(l, b);
      add_atom(l, c);

      l= new_clause(cl);
      add_atom(l, -a);
      add_atom(l, -b);
      add_atom(l, -c);

      closify(nd, cl, d, get_transition_tree_left_son(d, t));
      return closify(nd, cl, d, get_transition_tree_right_son(d, t));
    }

    case EQUAL : { /* a = (b = c) == (a + !b + !c) . (a + b + c) . (!a + !b + c) . (!a + b + !c) */
      unsigned long a = d->trans_tree[t].number;
      unsigned long b = d->trans_tree[get_transition_tree_left_son(d, t)].number;
      unsigned long c = d->trans_tree[get_transition_tree_right_son(d, t)].number;
      clause *l;

      l= new_clause(cl);
      add_atom(l, a);
      add_atom(l, -b);
      add_atom(l, -c);

      l= new_clause(cl);
      add_atom(l, a);
      add_atom(l, b);
      add_atom(l, c);

      l= new_clause(cl);
      add_atom(l, -a);
      add_atom(l, -b);
      add_atom(l, c);

      l= new_clause(cl);
      add_atom(l, -a);
      add_atom(l, b);
      add_atom(l, -c);

      closify(nd, cl, d, get_transition_tree_left_son(d, t));
      return closify(nd, cl, d, get_transition_tree_right_son(d, t));
    }

    case IMPLIES : { /* a = (b => c) == (a + b) . (a + !c) . (!a . !b . c) */
      unsigned long a = d->trans_tree[t].number;
      unsigned long b = d->trans_tree[get_transition_tree_left_son(d, t)].number;
      unsigned long c = d->trans_tree[get_transition_tree_right_son(d, t)].number;
      clause *l;

      l= new_clause(cl);
      add_atom(l, a);
      add_atom(l, b);

      l= new_clause(cl);
      add_atom(l, a);
      add_atom(l, -c);

      l= new_clause(cl);
      add_atom(l, -a);
      add_atom(l, -b);
      add_atom(l, c);

      closify(nd, cl, d, get_transition_tree_left_son(d, t));
      return closify(nd, cl, d, get_transition_tree_right_son(d, t));
    }

    default :
      fprintf(stderr, "trans.c:%d:closify: error : unhandled type %d\n", __LINE__, get_transition_tree_type(d, t));
      abort();
  }

#else

  switch(get_transition_tree_type(d, t)) {
    case VAR :
    case CTL_TRUE :
    case CTL_FALSE :
      d->trans_tree[t].added = 1;
      return;
    case NOT :
      closify(nd, cl, d, get_transition_tree_son(d, t));
      add_clause(nd, cl, d, t);
      return;
    case AND :
    case OR :
    case XOR :
    case EQUAL :
    case IMPLIES :
      closify(nd, cl, d, get_transition_tree_left_son(d, t));
      closify(nd, cl, d, get_transition_tree_right_son(d, t));
      add_clause(nd, cl, d, t);
      return;
    default :
      fprintf(stderr, "trans.c:%d:numberize: error : unhandled type %d\n", __LINE__, get_transition_tree_type(d, t));
      abort();
  }
#endif /* 0 */
}

/* returns the number of variables created */
unsigned long to_cnf(clauses *c, trans_data *d, transition_tree t)
{
  /* we want to avoid conflicts with our current trans_data, so a new one */
  trans_data new_trans;
  extern checker the_checker;

  if (init_trans(&new_trans)==-1) {
    fprintf(stderr, "out of memory\n");
    return 0;
  }

  print("cnf", stdout, "initial transition function :\n");
  if (verbose_on("cnf")) {
    print_modified_tree(d, &the_checker.variables, t, the_checker.vars);
    fprintf(stdout, "\n");
  }

  nn = 1;
  numberize(d, t);

  vars = getmem(nn * sizeof(char));
  vars_size = nn;

  memset(vars, 0, vars_size * sizeof(char));

  closify(&new_trans, c, d, t);

  freemem(vars);

  free_trans(&new_trans);

  return nn-1;
}

void dump_variable_names(trans_data *d, transition_tree t, variables *v, bdd_variables_set *s)
{
  if (d->trans_tree[t].added == 0)
    return;
  d->trans_tree[t].added = 0;

  switch(get_transition_tree_type(d, t)) {
    case CTL_TRUE :
      fprintf(stdout, "c %d = Strue[0]\n", d->trans_tree[t].number);
      return;
    case CTL_FALSE :
      fprintf(stdout, "c %d = Sfalse[0]\n", d->trans_tree[t].number);
      return;
    case VAR :
      fprintf(stdout, "c %d = V%d_%s[0]\n", d->trans_tree[t].number, get_transition_tree_variable_pos(d, t),
					    get_variable_name(v, s->list2[get_transition_tree_variable(d, t)]));
      return;
    case NOT :
      dump_variable_names(d, get_transition_tree_son(d, t), v, s);
      return;
    case AND :
    case OR :
    case XOR :
    case IMPLIES :
    case EQUAL :
      dump_variable_names(d, get_transition_tree_left_son(d, t), v, s);
      dump_variable_names(d, get_transition_tree_right_son(d, t), v, s);
      return;
  }
}
