/*
 * A mini transition relation postfix reader
 * ---
 * January 2002
 * to my angel
 */

#include "trans.h"

#include <string.h>

#include "vars.h"
#include "mem.h"
#include "checker.h"
#include "ctl.h"
#include "hash2.h"
#include "verbose.h"
#include "clauses.h"
#include "bmc.h"

static transition_tree new_tree(trans_data *d)
{
  if (d->size == d->maxsize)
    d->trans_tree = regetmem(d->trans_tree, sizeof(_transition_tree) * (d->maxsize += 32768));

  memset(&d->trans_tree[d->size], 0, sizeof(_transition_tree));

  d->size++;

  return d->size-1;
}

transition_tree new_transition_tree_var2(trans_data *d, variable v, int pos)
{
  transition_tree t;
  unsigned long h;

  h = hash2_get_new_slot(&d->ht, VAR, v, pos);
  if (d->ht.flag == HASH2_ALREADY_PRESENT) {
    transition_tree found = get_hash_node(&d->ht, h);
    print("hash", stdout, "hash already present : node %lu = var %u\n", found, v);
    return found;
  }

  /* handle data */
  t = new_tree(d);

  hash2_feed_slot(&d->ht, h, t, VAR, v, pos);

  set_transition_tree_type(d, t, VAR);
  set_transition_tree_variable(d, t, v);
  set_transition_tree_variable_pos(d, t, pos);

  return t;
}

transition_tree new_transition_tree_var(trans_data *d, variable v)
{
  return new_transition_tree_var2(d, v, -1);
}

transition_tree new_transition_tree_op(trans_data *d, int op, transition_tree l, transition_tree r)
{
  transition_tree t;
  unsigned long h;

  /* sanity checks */
  switch (op) {
    case OR :
    case XOR :
    case AND :
    case EQUAL :
    case IMPLIES :
      if (trans_undef(l) || trans_undef(r))
	return TRANS_UNDEF;
      break;
    case NOT :
      if (trans_undef(l) || !trans_undef(r))
	return TRANS_UNDEF;
      break;
    default : return TRANS_UNDEF;
  }

  /* handle hash table */
  switch (op) {
    case OR :
    case XOR :
    case AND :
    case EQUAL :
    case IMPLIES :
      h = hash2_get_new_slot(&d->ht, op, l, r);
      break;
    case NOT :
      h = hash2_get_new_slot(&d->ht, op, l, TRANS_UNDEF);
      break;
    default : /* one can't come here */ abort();
  }

  if (d->ht.flag == HASH2_ALREADY_PRESENT) {
    transition_tree found = get_hash_node(&d->ht, h);
    print("hash", stdout, "hash already present : node %lu = op %u args %lu %lu\n", found, op, l, r);
    return found;
  }

  /* handle data */
  t = new_tree(d);

  set_transition_tree_type(d, t, op);

  switch (op) {
    case OR :
    case XOR :
    case AND :
    case EQUAL :
    case IMPLIES :
      set_transition_tree_left_son(d, t, l);
      set_transition_tree_right_son(d, t, r);
      hash2_feed_slot(&d->ht, h, t, op, l, r);
      break;
    case NOT :
      set_transition_tree_son(d, t, l);
      hash2_feed_slot(&d->ht, h, t, op, l, TRANS_UNDEF);
      break;
  }

  return t;
}

transition_tree new_transition_tree_true(trans_data *d)
{
  return 1;
}

transition_tree new_transition_tree_false(trans_data *d)
{
  return 0;
}

void free_transition_tree(trans_data *d, transition_tree t)
{
  /* nothing to do ? */
}

int init_trans(trans_data *d)
{
#if 0
  transition_tree t;
  unsigned long h;
#endif

  memset(d, 0, sizeof(trans_data));

  init_hash2(&d->ht, HASH2_INIT_SIZE_BIG);

#if 0
  print("hash", stdout, "creating ctl_false\n");

  /* let's insert CTL_TRUE, CTL_FALSE */
  h = hash2_get_new_slot(&d->ht, CTL_FALSE, TRANS_UNDEF, TRANS_UNDEF);
  if (d->ht.flag != HASH2_NO_ERROR) abort();
  t = new_tree(d);
  hash2_feed_slot(&d->ht, h, t, CTL_FALSE, TRANS_UNDEF, TRANS_UNDEF);
  set_transition_tree_type(d, t, CTL_FALSE);

  print("hash", stdout, "creating ctl_true\n");

  h = hash2_get_new_slot(&d->ht, CTL_TRUE, TRANS_UNDEF, TRANS_UNDEF);
  if (d->ht.flag != HASH2_NO_ERROR) abort();
  t = new_tree(d);
  hash2_feed_slot(&d->ht, h, t, CTL_TRUE, TRANS_UNDEF, TRANS_UNDEF);
  set_transition_tree_type(d, t, CTL_TRUE);

  print("hash", stdout, "created both finished\n");
#endif

  return 0;
}

void free_trans(trans_data *d)
{
  print("hash", stdout, "trans.c:free_trans : ");

  free_hash2(&d->ht);

  if (d->maxsize)
    freemem(d->trans_tree);

  d->size = d->maxsize = 0;
  d->trans_tree = 0;
}
