/*
 * A mini transition relation postfix reader
 * ---
 * January 2002
 * to my angel
 */

#include "vars.h"

#include <string.h>
#ifdef DEBUG
#include <stdio.h>
#endif

#include "mem.h"
#include "hash.h"
#include "strings.h"

#define SIZEINC	16

static sid user_to_sid(variables *v, variable *var)
{
#ifdef DEBUG
  if (*var < 0 || *var >= v->size) {
    fprintf(stderr, "vars.c:user_to_sid (line %d): error : *var = %d\n", __LINE__, *var);
    abort();
  }
#endif

  return v->tab[*var];
}

/* return 0 if ok, -1 if pb */
int init_vars(variables *v)
{
  v->tab = 0;
  v->tab_assoc = 0;
  v->type = 0;
  v->size = v->maxsize = 0;

  return init_strings(&v->strings, (string_user_to_sid_function)user_to_sid, (user_context *)v, sizeof(variable));
}

void free_vars(variables *v)
{
  free_strings(&v->strings);

  if (v->maxsize) {
    freemem(v->tab);
    freemem(v->tab_assoc);
    freemem(v->type);
  }

  v->size = v->maxsize = 0;
  v->tab = 0;
  v->tab_assoc = 0;
  v->type = 0;
}

/* return the variable if it exists, NO_VARIABLE if not */
static inline variable find_variable(variables *v, char *s)
{
  sid sid;
  variable *var;

  if (string_search(&v->strings, s, &sid, (user_data **)&var) == -1)
    return NO_VARIABLE;

  return *var;
}

/* return NO_VARIABLE on error, the var if all ok */
variable new_variable(variables *v, char *s)
{
  sid sd;
  hid hd;
  variable d;

  if(string_new(&v->strings, s, &hd, &sd)==-1)
    return NO_VARIABLE;

  d = v->size;
  string_feed_slot(&v->strings, hd, (void *)&d);

  if (v->size == v->maxsize) {
    v->tab = regetmem(v->tab, sizeof(sid *) * (v->maxsize+=SIZEINC));
    v->tab_assoc = regetmem(v->tab_assoc, sizeof(void *) * v->maxsize);
    v->type = regetmem(v->type, sizeof(info_type) * v->maxsize);
  }

  v->tab_assoc[v->size] = 0;
  v->tab[v->size++] = sd;

  return d;
}

variable get_variable(variables *v, char *s)
{
  return find_variable(v, s);
}

char *get_variable_name(variables *v, variable x)
{
  if (x<0 || x>= v->size)
    return 0;

  return sid_to_string(&v->strings, v->tab[x]);
}

void *get_variable_assoc(variables *v, variable x)
{
  if (x<0 || x>= v->size)
    return 0;

  return v->tab_assoc[x];
}

void set_variable_assoc(variables *v, variable x, void *y)
{
  if (x<0 || x>= v->size)
    return;

  v->tab_assoc[x] = y;
}

void set_variable_type(variables *v, variable x, int type)
{
  if (x<0 || x>= v->size)
    return;

  v->type[x].type = type;
}

int get_variable_type(variables *v, variable x)
{
  if (x<0 || x>= v->size)
    return VARIABLE_TYPE_NO_TYPE;

  return v->type[x].type;
}

void set_variable_canon(variables *v, variable x, variable canon )
{
  if (x<0 || x>= v->size)
    return;

  v->type[x].canonik_variable = canon;
}

variable get_variable_canon(variables *v, variable x)
{
  if (x<0 || x>= v->size)
    return NO_VARIABLE;

  return v->type[x].canonik_variable;
}
